module Main
  ( main,
  )
where

import Control.Monad
import Control.Monad.State
import Data.ZSDD.Internal
import System.Exit (exitFailure)
import System.Random
import Prelude hiding (negate)

-- Hi and Lo destinations must be Top, Bot or a larger atom.
-- Hi never points to Bot
checkInvariants :: Prop d Int -> Diagram d Int IO Bool
checkInvariants Top = return True
checkInvariants Bot = return True
checkInvariants (ID i) = withNode i $ \n ->
  case hi n of
    Top -> checkInvariants (lo n)
    Bot -> return False
    ID j -> withNode j $ \m ->
      if atom n < atom m
        then liftM2 (&&) (checkInvariants $ lo n) (checkInvariants $ hi n)
        else return False

-- Generate a random operation
randOp :: Diagram d Int IO (Ops Int)
randOp = Diagram $ do
  (i, _, _, _) <- get
  if i == 0
    then Atomic <$> lift randomIO
    else do
      r <- lift (randomRIO (0, 4))
      case r :: Int of
        0 -> Atomic <$> lift randomIO
        1 -> Union <$> lift (randomRIO (0, i - 1)) <*> lift (randomRIO (0, i - 1))
        2 -> Intersect <$> lift (randomRIO (0, i - 1)) <*> lift (randomRIO (0, i - 1))
        3 -> Negate <$> lift (randomRIO (0, i - 1)) <*> lift randomIO
        4 -> Assign <$> lift (randomRIO (0, i - 1)) <*> lift randomIO <*> lift randomIO

-- Get a random proposition after applying n operations
randProp :: Int -> Diagram d Int IO (Prop d Int)
randProp n = do
  ops <- replicateM n randOp
  props <- mapM runOp ops
  return (last props)

-- Apply an operation
runOp :: Ops Int -> Diagram d Int IO (Prop d Int)
runOp (Atomic a) = atomic a
runOp (Union p q) = ID p `union` ID q
runOp (Intersect p q) = ID p `intersect` ID q
runOp (Negate p a) = negate (ID p) a
runOp (Assign p a b) = assign (ID p) a b

main :: IO ()
main =
  replicateM_ 100 $ do
    c <- runDiagram (randProp 10 >>= checkInvariants)
    unless c exitFailure
